#!/usr/bin/php
<?php
/**
 * PhotoTools
 *
 * A toolset for managing photos and videos from a digital camera 
 * or storage medium
 *
 * Copyright (C) 2010 Roland Eckert <mail@rolandeckert.com>
 *
 * This program is free software; you can redistribute it and/or modify it 
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 3 of the License, or (at your option)
 * any later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or 
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, see <http://www.gnu.org/licenses/>.
 *
 *
 * TODOs: 
 * - remove temporary input folder after move if empty
 * - check available disk space before import, duplicate or move
 *
 */

require('./getid3/getid3/getid3.php');

$config = array(
  'DEFAULT_MEDIA_PATH'=>'/media/',
  'DEFAULT_TEMP_PATH' => $_SERVER['HOME'].'/tmp/',
  'SUPPORTED_FILE_EXTENSIONS' => array('jpg', 'avi', 'mov'),
  'DUPLICATE_EXTENSION' => '_duplicate'
);

class PhotoTools {

  var $c = array();
  var $dryrun = false;

  function __construct($c) {
    $this->c = $c;
  }

  function m($psMsg, $uline=false) {
    echo (($uline)?"\n":'')
      .$psMsg."\n"
      .(($uline)?str_repeat('-', strlen($psMsg))."\n":'');
  }

  function usage() {
    $n = basename($_SERVER['PHP_SELF']);
    $f = implode(', ', $this->c['SUPPORTED_FILE_EXTENSIONS']);
    echo <<<EOT

PhotoTools is a toolset for managing photos and videos from a digital camera 
or storage medium.

Copyright (C) 2010 Roland Eckert <mail@rolandeckert.com>

  This program comes with ABSOLUTELY NO WARRANTY.
  This is free software, and you are welcome to redistribute it
  under certain conditions. See COPYING for details.

Usage:

  ./$n [COMMAND]... [OPTION]...

  ./$n delete <source> [go]
  ./$n duplicate <source>
  ./$n import <source>
  ./$n index <source>
  ./$n list
  ./$n meta <file>
  ./$n move <source> <target> [go]
  ./$n strip <source> <master> [go]

Commands:

  list      Lists all available folders from the default mount point ({$this->c['DEFAULT_MEDIA_PATH']})

  delete    Delete all files from the specified source folder. Dry-run mode (no 
            action taken) if "go" isn't added to end of command line
  
  duplicate Duplicates the given source folder in the same location, 
            appending a default extension ({$this->c['DUPLICATE_EXTENSION']})

  import    imports all supported file types recursively from the given source
            folder and copies them to the default temp folder ({$this->c['DEFAULT_TEMP_PATH']})
            Supported file types are: $f
            
  index     create contact sheet containing all images as thumbnails

  meta      Read meta data from file, e.g. EXIF data

  move      Moves the images from the specified source folder to the target
            folder, using EXIF, ID3 or file date information for a suitable 
            folder name. Dry-run mode (no action taken) if "go" isn't added
            to end of command line

  strip     Removes all files from the given source folder which are present in
            the given master folder. Dry-run mode (no action taken) if "go"
            isn't added to end of command line

Please report bugs to Roland Eckert <mail@rolandeckert.com>.

EOT;
  }
  /* ' workaround for stupid highlighting bug in my IDE :-) */

  function deleteFiles() {

    $target = realpath($this->location);
    $this->m('Deleting folder contents '.$target.'/*');

    if (!$this->dryrun && $target) {
      $this->m('rm -rf '.$target.'/* 2>&1');
      $output = shell_exec('rm -rf '.$target.'/* 2>&1');
      if ($output) {
        $this->m("Error: Removing folder gave output:\n".$output);
        exit;
      }     
    }
  }

  function duplicateFiles() {

    $target = rtrim($this->location,'/').$this->c['DUPLICATE_EXTENSION'];
    if (file_exists($target)) {
      $this->m('Error: Path for duplicate exists');
      exit;
    }

    // TODO: check space with disk_free_space()

    $this->m('Duplicating folder '.$this->location.' to '.$target);
    $output = shell_exec('cp -a '.escapeshellarg($this->location).' '.escapeshellarg($target).' 2>&1');
    if ($output) {
      $this->m("Error: Duplicating folder gave output:\n".$output);
      exit;
    }     

  }

  function importFiles() {

    $this->m('Importing from folder '.$this->location);

    if(!file_exists($this->c['DEFAULT_TEMP_PATH']) || !is_writeable($this->c['DEFAULT_TEMP_PATH'])) {
      $this->m('Error: DEFAULT_TEMP_PATH => '.$this->c['DEFAULT_TEMP_PATH'].' is not writeable');
      exit;
    }

    // TODO: check space with disk_free_space()

    $folder = $this->c['DEFAULT_TEMP_PATH'].basename($this->location).'-'.date('Y-m-d');
    $success = false;
    if (!file_exists($folder)) {
      mkdir($folder);
    } else {
      for($c=1; $c<4; $c++) {
        if(!file_exists($folder.'_'.$c)) {
          $folder.= '_'.$c;
          mkdir($folder);
          $success = true;
          break;
        }
      }
      if(!$success) {
        $this->m('Error creating unique folder for import');
        exit;
      }
    }

    $this->m('Importing to folder: '.$folder);


    $bytestotal = 0;
    $nbfiles = 0;

    $iterator = new RecursiveDirectoryIterator($this->location);
    foreach (new RecursiveIteratorIterator($iterator) as $filename=>$fileobj) {

      $basename = $fileobj->getBasename();
      $extension = strtolower(substr($basename, strrpos($basename, '.')+1));

      # skip dot files      
      if ($basename=='.' || $basename=='..') {
        continue;
      }        

      $this->m('Processing file: '.$filename);

      if (!in_array($extension, $this->c['SUPPORTED_FILE_EXTENSIONS'])) {
        $this->m('Unsupported file: '.$filename);
        continue;
      }

      // copy file
      $target = $folder.'/'.$basename;
      if (!file_exists($target)) {

        $this->m('Copying file '.$filename.' to '.$target);

        // copy() does not preserve filetimes
        $output = shell_exec('cp --preserve=timestamps '.escapeshellarg($filename).' '.escapeshellarg($target).' 2>&1');
        if($output) {
          $this->m('Error: Could not copy file '.$filename.' to '.$target);
          $this->m("Copying file gave output:\n".$output);
          exit;
        }

        $filesize = $fileobj->getSize();
        $bytestotal+= $filesize;
        $nbfiles++;

      } else {
        $this->m('Warning: file '.$target.' exists, skipping!');
      }

    }

    $bytestotal = number_format($bytestotal);
    $this->m("Copied $nbfiles files, $bytestotal bytes");

  }

  function createIndex() {

    $path = realpath($this->location);
    $files = glob($path."/{*.jpg,*.JPG}", GLOB_BRACE);
    $outfile = $this->c['DEFAULT_TEMP_PATH'].'index_'.str_replace('/', '_', $path).'.jpg';

    $cmd = "/usr/bin/montage ".implode(' ', $files)." \
            -label '%f' -tile 10x -frame 3 -shadow \
            -title '$path' \
            -geometry '60x60+4+4>' -background lightblue \
            $outfile 2>&1";

    $output=array();
    exec($cmd, $output, $return_value);

    if ($return_value!==0) {
      $this->m('Index creation failed');
      $this->m(implode("\n", $output));
      exit;
    }

    $this->m('Index has been created: '.$outfile);
  }

  function listFiles() {
    $this->m($this->c['DEFAULT_MEDIA_PATH'].' contents', 1);
    system('ls -lartsh '.$this->c['DEFAULT_MEDIA_PATH']);
    $this->m('folder sizes', 1);
    system('du -h --max-depth=1 '.$this->c['DEFAULT_MEDIA_PATH']);
  }

  function showMeta() {

    $data = $this->_getMeta($this->location);
    if ($data===false) {
      $this->m('Warning: No meta data found in '.$this->location);
      exit;
    }
    $s = print_r($data, 1);
    echo $this->_clispecialchars($s);
  }

  function moveFiles() {

    $this->m('Moving files from folder '.$this->location.' to '.$this->target);

    if(!is_writeable($this->target)) {
      $this->m('Error: Specified target is not writeable');
      exit;
    }

    $bytestotal = 0;
    $nbfiles = 0;

    $iterator = new RecursiveDirectoryIterator($this->location);
    foreach (new RecursiveIteratorIterator($iterator) as $filename=>$fileobj) {

      $basename = $fileobj->getBasename();
      $extension = strtolower(substr($basename, strrpos($basename, '.')+1));

      # skip dot files      
      if ($basename=='.' || $basename=='..') {
        continue;
      }        

      $this->m('Processing file: '.$filename);

      if (!in_array($extension, $this->c['SUPPORTED_FILE_EXTENSIONS'])) {
        $this->m('Unsupported file: '.$filename);
        continue;
      }

      // try to read creation date from EXIF data or other metadata, otherwise use mtime

      $ts = false;
      $date = false;

      $meta = $this->_getMeta($filename);

      switch ($extension) {
      
        case 'avi':
          if (isset($meta['riff']['AVI ']['hdrl']['IDIT'][0]['data'])) {
            $date = trim($meta['riff']['AVI ']['hdrl']['IDIT'][0]['data']);
          }
          if ($date) {
            $ts = strtotime($date);
          }

        break;

        case 'jpg':
          if (isset($meta['IFD0']['DateTime'])) {
            $date = trim($meta['IFD0']['DateTime']);
          } elseif (isset($meta['EXIF']['DateTimeOriginal'])) {
            $date = trim($meta['EXIF']['DateTimeOriginal']);
          }
          if ($date) {
            $ts = strtotime($date);
          }
        break;

        case 'mov':
          if (isset($meta['quicktime']['moov']['subatoms']['0']['creation_time_unix'])) {
            $ts = trim($meta['quicktime']['moov']['subatoms']['0']['creation_time_unix']);
          }
        break;

      }

      if(!$ts || $ts<10000000) {
        $this->m('Creation date not present or not plausible, using filemtime');
        $ts = filemtime($filename);
      }

      $this->m('Using creation date: '.date('r', $ts));
      
      // target folder
      $datestr = date('Y', $ts).'-'.date('m', $ts);

      $targetfolder = $this->target.'/'.$datestr;
      if (!file_exists($targetfolder)) {
        $this->m('Creating folder '.$targetfolder);
        if (!$this->dryrun) {
          if (!mkdir($targetfolder)) {
            $this->m('Error: Could not create folder '.$targetfolder);
            exit;
          }
        }
      }

      // move file to target folder
      $targetfile = $targetfolder.'/'.$basename;
      if (file_exists($targetfile)) {
        // filename already exists! do checksum check
        $this->m('Warning: Target file '.$targetfile.' exists, checking MD5 sum of source and target');
        $shash = md5_file($filename);
        $thash = md5_file($targetfile);
        if ($shash!=$thash) {
          $this->m('Warning: Target file '.$targetfile.' exists with different content, skipping');
          continue;
        } 
        if (!$this->overwrite) {
          $this->m('Overwrite is not set, skipping identical file');
          continue;
        }
        $this->m('Overwrite is set, moving identical file over target');
      }

      if ($this->dryrun) {
        $this->m('(Dry-run) Moving '.$filename.' to '.$targetfile);
        continue;
      }

      $filesize = $fileobj->getSize();
      $bytestotal+= $filesize;
      $nbfiles++;

      $this->m('Moving '.$filename.' to '.$targetfile);

      // rename() does not preserve filetimes
      $output = shell_exec('mv '.escapeshellarg($filename).' '.escapeshellarg($targetfile).' 2>&1');
      if($output) {
        $this->m('Error: Could not move file '.$filename.' to '.$targetfile);
        $this->m("Moving file gave output\n".$output);
        exit;
      }

    }

    $bytestotal = number_format($bytestotal);
    $this->m("Moved $nbfiles files, $bytestotal bytes");

  }

  function stripFiles() {
    
    $iterator = new RecursiveDirectoryIterator($this->location);
    foreach (new RecursiveIteratorIterator($iterator) as $filename=>$fileobj) {

      $basename = $fileobj->getBasename();

      # skip dot files      
      if ($basename=='.' || $basename=='..') {
        continue;
      }

      $this->m('Processing file: '.$filename);

      // delete file if present in master folder
      $master = $this->master.'/'.$basename;
      if (file_exists($master)) {
        $this->m('Deleting '.$filename);
        if (!$this->dryrun) {
          if (!unlink($filename)) {
            $this->m('Error: Could not delete file '.$filename);
            exit;
          }
        }
      } else {
        $this->m('Keeping '.$filename);
      }
    }
  }

  function _clispecialchars($psS) {
    return preg_replace('/[^(\x09|\x0a|(\x20-\x7F))]*/','', $psS);
  }

  /**
   * read meta data from file. Currently JPG, AVI and MOV are supported
   */
  function _getMeta($psFile) {

    $basename = basename($psFile);
    $extension = strtolower(substr($basename, strrpos($basename, '.')+1));

    $data = null;
    switch ($extension) {

      case 'jpg':
        $data = @exif_read_data($psFile, 0, true);
      break;

      case 'avi':
      case 'mov':
        $getID3 = new getID3;
        $data = $getID3->analyze($psFile);
        getid3_lib::CopyTagsToComments($data);
      break;

      default:
        $this->m('Warning: Unsupported file type '.$extension);
    }
    return $data;
  }

}


if (!isset($argc) || !isset($argv)) {
  $argc = $_SERVER['argc'];
  $argv = $_SERVER['argv'];
}

$o = new PhotoTools($config);
 
if ($argc==1) {

  $o->usage();

} else {

  $o->command = $argv[1];
  switch ($o->command) {

    // delete folder contents on disk
    case 'delete':

      if (!isset($argv[2]) || !is_dir($argv[2])) {
        $o->m('Error: Location not specified or not a directory');
        exit;
      }
      if (!isset($argv[3]) || $argv[3]!='go') {
        $o->dryrun = true;
        $o->m('Dry-run mode. Add \'go\' to actually delete the files');
      }

      $o->location = $argv[2];       
      $o->deleteFiles();
      
    break;

    // duplicate a folder on disk
    case 'duplicate':

      if (!isset($argv[2]) || !is_dir($argv[2])) {
        $o->m('Error: Location not specified or not a directory');
        exit;
      }

      $o->location = $argv[2];       
      $o->duplicateFiles();
      
    break;

    // recursively import files from a folder, i.e. storage medium 
    case 'import':

      if (!isset($argv[2]) || !is_dir($argv[2])) {
        $o->m('Error: Location not specified or not a directory');
        exit;
      }

      $o->location = $argv[2];       
      $o->importFiles();
      
    break;

    // create an index sheet with all files from one folder
    case 'index':

      if (!isset($argv[2]) || !is_dir($argv[2])) {
        $o->m('Error: Location not specified or not a directory');
        exit;
      }

      $o->location = $argv[2];       
      $o->createIndex();
      
    break;

    // list folders and summaries from DEFAULT_MEDIA_PATH    
    case 'list':
      $o->listFiles();
    break;

    // read meta data (e.g. EXIF) from file    
    case 'meta':

      if (!isset($argv[2]) || !file_exists($argv[2])) {
        $o->m('Error: Specified file doesn\'t exist');
        exit;
      }

      $o->location = $argv[2];     
      $o->showMeta();
      
    break;

    // move files from one to another folder, using a suitable (date) folder name
    case 'move':

      if (!isset($argv[2]) || !is_dir($argv[2])) {
        $o->m('Error: Source location not specified or not a directory');
        exit;
      }
      if (!isset($argv[3]) || !is_dir($argv[3])) {
        $o->m('Error: Target location not specified or not a directory');
        exit;
      }
      if (!isset($argv[4]) || $argv[4]!='go') {
        $o->dryrun = true;
        $o->m('Dry-run mode. Add \'go\' to actually move the files');
      }

      $o->location = $argv[2];
      $o->target = $argv[3];
      $o->overwrite = true;
      $o->moveFiles();
      
    break;

    // 'strip' a folder, i.e. delete all files which are present in another folder    
    case 'strip':

      if (!isset($argv[2]) || !is_dir($argv[2])) {
        $o->m('Error: Location not specified or not a directory');
        exit;
      }
      if (!isset($argv[3]) || !is_dir($argv[3])) {
        $o->m('Error: Master not specified or not a directory');
        exit;
      }
      if (!isset($argv[4]) || $argv[4]!='go') {
        $o->dryrun = true;
        $o->m('Dry-run mode. Add \'go\' to actually strip the files from '.$argv[2]);
      }

      $o->location = $argv[2];     
      $o->master = $argv[3];     
      $o->stripFiles();
      
    break;
    
    default:
      $o->m('Error: unknown command '.$o->command);
      $o->usage();
  }
}

?>
